#! /usr/bin/python -O
# -*- coding: utf-8 -*-
#
# Simple GNU find replacement.
# Copyright (C) 2014 MiKael NAVARRO
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

"""Simple GNU find replacement.
Copyright (C) 2014 MiKael NAVARRO

"shell script"-like GNU find utility in Python.

"""

# Specific variables for pydoc
__author__ = "MiKael Navarro <klnavarro@gmail.com>"
__date__ = "Fri Feb 07 2014"
__version__ = "0.1.0"
__credits__ = """Thanks to Muharem Hrnjadovic for his 'scriptutil.py' module
(http://muharem.wordpress.com/2007/05/16/python-file-find-grep-and-in-line-replace-tools-part-1/)
and Pádraig Brady for his 'ls.py' example (http://www.pixelbeat.org/talks/python/ls.py.html)."""

# Include directives
import sys, os, re
from scriptutil import ffind, printr
import time, stat

# Global variables
USE_WILDCARDS = True
RELATIVE = True
NOW = int(time.time())
RECENT = NOW - (6*30*24*60*60) # 6 months ago
COLOURS={"default":"",
         "blue":   "\x1b[01;34m",
         "cyan":   "\x1b[01;36m",
         "green":  "\x1b[01;32m",
         "red":    "\x1b[01;05;37;41m"}

#
# Following from Python cookbook, #475186.
#
def has_colours(stream):
    """Determine if current terminal has colors support?"""

    if not hasattr(stream, "isatty"):
        return False
    if not stream.isatty():
        return False  # auto color only on TTYs
    try:
        import curses
        curses.setupterm()
        return curses.tigetnum("colors") > 2
    except:  # guess false in case of error
        return False

HAS_COLOURS = has_colours(sys.stdout)


#
# Function to get filename info from mode.
#
def get_mode_info(filename, mode):
    """Get info from filename mode."""

    perms = "-"
    colour = "default"
    link = ""

    if stat.S_ISDIR(mode):
        perms = "d"
        colour = "blue"
    elif stat.S_ISLNK(mode):
        perms = "l"
        colour = "cyan"
        link = os.readlink(filename)
        if not os.path.exists(filename):
            colour = "red"
    elif stat.S_ISREG(mode):
        if mode & (stat.S_IXGRP|stat.S_IXUSR|stat.S_IXOTH):
            colour = "green"
    mode = stat.S_IMODE(mode)
    for who in "USR", "GRP", "OTH":
        for what in "R", "W", "X":
            # Lookup attribute at runtime using getattr
            if mode & getattr(stat, "S_I"+what+who):
                perms = perms + what.lower()
            else:
                perms = perms + "-"

    # Return multiple bits of info in a tuple
    return (perms, colour, link)


#
# Main entry point.
#
def main(options, dirs):
    """Pass/interprets command line arguments to 'scriptutil.ffind' function."""

    if USE_WILDCARDS:
        # Unix shell-style wildcards
        wildcards = [options.pattern,]
        filters = []  # functions to be applied to the file paths found
    else:  # use regexp
        wildcards = []
        filters = [options.pattern and re.compile(options.pattern).search or None,]
    if __debug__: print "wildcards=", wildcards

    if options.type == 'f':  # regular file
        filters.append(lambda s: os.path.isfile(s),)
    if options.type == 'l':  # symlink
        filters.append(lambda s: os.path.islink(s),)

    if options.prune:  # ignore given pattern in the file paths found
        #filters.append(lambda s: s.find(options.prune) == -1,)
        filters.append(lambda s: not re.search(re.compile(options.prune), s),)
    if __debug__: print "filters=", filters

    # Call scriptutil.ffind...
    files = ffind(dirs[0],
                  shellglobs=wildcards,
                  namefs=filters,
                  relative=options.relative)
    if __debug__: print "files=", files

    # Print results
    if options.print0:
        for f in sorted(files):
            print(f),

    elif options.ls:
        for f in sorted(files):
            try:
                stat_info = os.lstat(f)
            except:
                print >> sys.stderr, "%s: error calling os.lstat!" % f
                continue

            # Get permissions...
            perms, colour, link = get_mode_info(f, stat_info.st_mode)

            # Number of hard links
            nlink = "%4d" % stat_info.st_nlink

            # User id of owner
            try:
                from pwd import getpwuid
                name = "%-8s" % getpwuid(stat_info.st_uid)[0]
            except:
                name = "%-8s" % stat_info.st_uid
            # Group id of owner
            try:
                from grp import getgrgid
                group = "%-8s" % getgrgid(stat_info.st_gid)[0]
            except:
                group = "%-8s" % stat_info.st_gid

            # Size of file, in bytes
            size = "%8d" % stat_info.st_size

            # Get time stamp of file (most recent content modification)
            ts = stat_info.st_mtime
            if (ts < RECENT) or (ts > NOW):
                time_fmt = "%b %d  %Y"
            else:
                time_fmt = "%b %d %H:%M"
            time_str = time.strftime(time_fmt, time.gmtime(ts))

            # Write the result
            print "%s %s %s %s %s %s " % (perms, nlink, name, group, size, time_str),
            if COLOURS[colour] and HAS_COLOURS:
                print COLOURS[colour] + f + "\x1b[00m",
            else:
                print f,
            if link:
                print "->",
            print link

    else:  #options.print1:
        printr(files)


#
# External entry point.
#
if __name__ == "__main__":
    # Get options
    from optparse import OptionParser, OptionGroup
    
    OptionParser.format_epilog = lambda self, formatter: self.epilog

    parser = OptionParser(usage="python -O %prog [options] dir1 [dir2 ...]", 
                          version=__version__,
                          description="Simple GNU find replacement.",
                          epilog="""
Examples:
  $ pyfind --type f --name "*.txt" --fullpath
  $ pyfind --type f --name "*.py" --prune ".bzr"
""")

    tests = OptionGroup(parser, "Tests", "")
    tests.add_option("--type",
                     action="store", dest="type",
                     type="string", default="f",
                     help="file is of type 'f' (regular file) or 'l' (symlink)")
    tests.add_option("--name",
                     action="store", dest="pattern",
                     type="string", default="*",
                     help="base of file name matches shell pattern PATTERN")
    tests.add_option("--prune",
                     action="store", dest="prune",
                     type="string", default="",
                     help="to ignore PRUNE pattern in the file paths found")
    parser.add_option_group(tests)

    actions = OptionGroup(parser, "Actions", "")
    actions.add_option("--print",
                       action="store_true", dest="print1", 
                       default=True,
                       help="print the file name on the standard output, followed by a newline (default behaviour)")
    actions.add_option("--print0",
                       action="store_true", dest="print0", 
                       default=False,
                       help="print the file name followed by a null character (instead of the newline character)")
    actions.add_option("--fullpath",
                       action="store_false", dest="relative", 
                       default=True,
                       help="list file with full path")
    actions.add_option("-l", "--ls",
                       action="store_true", dest="ls", 
                       default=False,
                       help="list current file in 'ls -dils' format on standard output")
    parser.add_option_group(actions)

    (options, args) = parser.parse_args()

    # Needed argument
    if not args:
        parser.error("missing directories paths")

    # Check options
    if options.type and options.type not in ['f', 'l']:
        parser.error("incorrect type format")

    # Process start here
    main(options, args)
